package Stanford::PKObj ;

## no critic (Subroutines::RequireArgUnpacking);

use warnings ;
use strict ;
use autodie ;

use Stanford::Orange::Util qw/
  trim
  tempfile
  spew
  run_command
  run_command_improved
  slurp
/;

use MIME::Base64 ;
use Carp ;
use Data::Dumper ;
use File::Temp ;
use IPC::Run qw( run ) ;

our $VERSION = '0.02' ;

# We want an ordering of types in order of the most commonly to the least
# commonly encountered (helps when guessing).
my @ALLOWED_TYPES = qw
(
  CERTIFICATE
  REQUEST
  PRIVKEY
  PUBKEY
  CRL
) ;


# Hashify @ALLOWED_TYPES:
my %ALLOWED_TYPES = () ;
foreach my $type (@ALLOWED_TYPES) {
  $ALLOWED_TYPES{$type} = 1 ;
}

my %TYPE_TO_OPENSSL_PREFIX =
(
  'REQUEST'     => 'req',
  'CERTIFICATE' => 'x509',
  'PRIVKEY'     => 'pkey',
  'PUBKEY'      => 'pkey',
  'CRL'         => 'crl',
) ;

my %ALLOWED_ENCODINGS =
(
  'PEM' => 1,
) ;

our %TEXT_TO_TYPE =
(
  'cert'        => 'CERTIFICATE',
  'certificate' => 'CERTIFICATE',
  'privkey'     => 'PRIVKEY',
  'pubkey'      => 'PUBKEY',
  'csr'         => 'REQUEST',
  'req'         => 'REQUEST',
  'request'     => 'REQUEST',
  'crl'         => 'CRL',
) ;

my %TYPE_TO_HEADER =
(
  $TEXT_TO_TYPE{'cert'}      => '-----BEGIN CERTIFICATE-----',
  $TEXT_TO_TYPE{'privkey'}   => '-----BEGIN RSA PRIVATE KEY-----',
  $TEXT_TO_TYPE{'request'}   => '-----BEGIN CERTIFICATE REQUEST-----',
  $TEXT_TO_TYPE{'crl'}       => '-----BEGIN X509 CRL-----',
  $TEXT_TO_TYPE{'pubkey'}    => '-----BEGIN PUBLIC KEY-----',
) ;

my %HEADER_TO_TYPE = reverse %TYPE_TO_HEADER ;

# Add more headers we recognize
$HEADER_TO_TYPE{'-----BEGIN NEW CERTIFICATE REQUEST-----'} = $TEXT_TO_TYPE{'request'} ;
$HEADER_TO_TYPE{'-----BEGIN PRIVATE KEY-----'}             = $TEXT_TO_TYPE{'privkey'} ;
$HEADER_TO_TYPE{'-----BEGIN EC PRIVATE KEY-----'}          = $TEXT_TO_TYPE{'privkey'} ;


my %TYPE_TO_FOOTER =
(
  $TEXT_TO_TYPE{'cert'}      => '-----END CERTIFICATE-----',
  $TEXT_TO_TYPE{'privkey'}   => '-----END RSA PRIVATE KEY-----',
  $TEXT_TO_TYPE{'request'}   => '-----END CERTIFICATE REQUEST-----',
  $TEXT_TO_TYPE{'crl'}       => '-----END X509 CRL-----',
) ;

my %FOOTER_TO_TYPE = reverse %TYPE_TO_FOOTER ;

# Add one more footer that we recognize
$FOOTER_TO_TYPE{'-----END NEW CERTIFICATE REQUEST-----'} = $TEXT_TO_TYPE{'request'} ;


sub new {
    my ($proto, %args) = @_ ;
    my $self = {} ;
    my $class = ref($proto) || $proto ;

    $self->{'type'}     = $args{'type'} || undef ;
    $self->{'encoding'} = $args{'encoding'} || undef ;
    $self->{'contents'} = $args{'contents'} || undef ;
    $self->{'file'}     = $args{'file'} || undef ;

    bless($self, $class) ;

    $self->initialize() ;

    return $self ;
}

sub initialize {
    my $self = shift ;

    my $file = $self->get_file() ;
    if ($file) {
        $self->set_contents(slurp($file)) ;
    }

    my $contents = $self->get_contents() ;
    if ($contents) {
        my @objects = get_objects_from_string($contents) ;
        if (scalar(@objects) <= 0){
            if ($file) {
                croak "no valid objects in file '$file'" ;
            } else {
                croak 'no valid objects for supplied contents' ;
            }
        }
        my $new_pkobj = $objects[0] ;
        $self->set_type($new_pkobj->get_type()) ;
        $self->set_encoding($new_pkobj->get_encoding()) ;
    }

    return ;
}

sub to_s {
    my $self = shift ;

    my $type     = $self->get_type() ;
    my $encoding = $self->get_encoding() ;
    my $md5      = $self->get_md5() ;

    return <<"EOT"
type:     $type
encoding: $encoding
md5:      $md5
EOT
}

sub get_type {
    my $self = shift ;
    return $self->{'type'} ;
}

sub set_type{
    my $self = shift ;
    my ($type) = @_ ;

    # Allow setting type to undef.
    if (defined($type) && !$ALLOWED_TYPES{$type}) {
        croak "type '$type' not recognized" ;
    }

    $self->{'type'} = $type ;
    return ;
}

sub get_contents {
    my $self = shift ;
    return $self->{'contents'} ;
}

sub set_contents {
    my $self = shift ;
    my ($contents) = @_ ;
    $self->{'contents'} = $contents ;
    return ;
}

sub get_file {
    my $self = shift ;
    return $self->{'file'} ;
}

sub set_file {
    my $self = shift ;
    my ($file) = @_ ;
    $self->{'file'} = $file ;
    return ;
}

sub get_encoding {
    my $self = shift ;
    return $self->{'encoding'} ;
}

sub set_encoding {
    my $self = shift ;
    my ($encoding) = @_ ;
    if (!$ALLOWED_ENCODINGS{$encoding}) {
        croak "encoding '$encoding' not recognized" ;
    }

    $self->{'encoding'} = $encoding ;
    return ;
}

sub get_md5_old {
    my $self = shift ;
    my $type = $self->get_type() ;
    my $openssl_prefix = $TYPE_TO_OPENSSL_PREFIX{$type} ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;
    close($fh) ;

    spew($self->get_contents(), $tempfile) ;

    # public keys need an extra option.
    my $pubin_option = q{} ;
    if ($self->is_pubkey()) {
        $pubin_option = '-pubin' ;
    }

    # my $cmd = "openssl $openssl_prefix $pubin_option -noout -modulus -in $tempfile | openssl md5" ;
    my @cmd_modulus = (
        'openssl',
        $openssl_prefix,
        ) ;
    if ($pubin_option)  {
        @cmd_modulus = (@cmd_modulus, $pubin_option) ;
    }
    @cmd_modulus = (
        @cmd_modulus,
        '-noout',
        '-modulus',
        '-in', $tempfile,
        ) ;

    ## no critic (CodeLayout::ProhibitQuotedWordLists);
    my @cmd_md5 = (
        'openssl',
        'md5',
        ) ;

    my ($stdout, $stderr, $exit_value) ;
    run \@cmd_modulus, q{|}, \@cmd_md5, \$stdout, \$stderr ;

    if ($stderr) {
        croak "failed to run command '@cmd_modulus': $stderr" ;
    }

    $stdout = lc($stdout) ;

    # If there is a prefix of "(stdin)=" remove it.
    $stdout =~ s{^\(stdin\)=\s*}{}xsm ;

    my $md5 ;
    if ($stdout =~ m{([[:xdigit:]]{32})}xsm) {
        $md5 = $1 ;
    }
    else {
        croak "no MD5 hash output?!? ($stdout)" ;
    }

    return $md5 ;
}

# Returns the PEM-encoded public key
sub get_public_key { 
    my $self = shift ;
    my $type = $self->get_type() ;

    my ($stdin, $stdout, $stderr, $exit_value) ;

    # The public key is a special. If it is not a valid
    # public key raise an exception. If it _is_ valid, we can proceed like
    # all the other types.
    if ($self->is_pubkey() && (!$self->is_valid_public_key())) {
        croak 'this object is not a valid public key' ;
    }

    # Need a few temporary files.
    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;
    close($fh) ;

    spew($self->get_contents(), $tempfile) ;

    # Command to get public key in PEM format:
    # - CSR:
    #     openssl req  -pubkey -in file.csr -out out.pubkey -noout
    # - CRT:
    #     openssl x509 -pubkey -in file.pem -out out.pubkey -noout
    # - PRIVATE KEY:
    #     openssl pkey -pubout -in file.key -out out.pubkey
    # - PUBLIC KEY:
    #     cat file.pubkey

    my %type_to_cmd = (
        'REQUEST' =>
        [
         'openssl',
         'req',
         '-pubkey',
         '-in',  $tempfile,
         '-noout',
        ],
        'CERTIFICATE' =>
        [
         'openssl',
         'x509',
         '-pubkey',
         '-in',  $tempfile,
         '-noout',
        ],
        'PRIVKEY' =>
        [
         'openssl',
         'pkey',
         '-pubout',
         '-in',  $tempfile,
        ],
        'PUBKEY' =>
        [
         'cat',
         $tempfile,
        ],
    ) ;

    if (!exists($type_to_cmd{$type})) {
        croak 'do not know how to find md5 for this type of object' ;
    }

    my $cmd_aref = $type_to_cmd{$type} ;

    # Run this command.
    run $cmd_aref, \$stdin, \$stdout, \$stderr ;
    if ($stderr) {
        my $cmdx = join(q{ }, $cmd_aref) ;
        croak "failed to run command '$cmdx': $stderr" ;
    }

    return $stdout ;
}

# Take the md5 checksum of the public key in PEM format:
# -----BEGIN PUBLIC KEY-----
# MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEWAUGNqUwg+0gYg309GWlQY8hBIsw
# u4fIb2wgaO+cJWgxNUfOHCJH8DEYwD3aXFawRKlVCowiZg+epNgT0RIStQ==
# -----END PUBLIC KEY-----
sub get_md5 {
    my $self = shift ;
    my $type = $self->get_type() ;
    my $openssl_prefix = $TYPE_TO_OPENSSL_PREFIX{$type} ;

    my ($stdin, $stdout, $stderr, $exit_value) ;

    my $public_key = $self->get_public_key() ;

    # Make a PUBKEY object from $public_key.
    my $pubkey = Stanford::PKObj->new() ;
    $pubkey->set_type('PUBKEY') ;
    $pubkey->set_encoding('PEM') ;
    $pubkey->set_contents($public_key) ;

    # Put the DER version of $pubkey into the file $tempfile_der.
    my $fh_der          = File::Temp->new(UNLINK => 1) ;
    my $tempfile_der    = $fh_der->filename ;
    close($fh_der) ;

    $pubkey->to_der_file($tempfile_der) ;

    # Now get the md5 version of this DER file.

    my @cmd_md5 = (
        'openssl',
        'md5',
        $tempfile_der,
        ) ;

    # Run this command.
    run \@cmd_md5, \$stdin, \$stdout, \$stderr ;
    if ($stderr) {
        my $cmdx = join(q{ }, @cmd_md5) ;
        croak "failed to run command '$cmdx': $stderr" ;
    }

    $stdout = lc($stdout) ;

    # If there is a prefix of "(stdin)=" remove it.
    $stdout =~ s{^\(stdin\)=\s*}{}xsm ;

    my $md5 ;
    if ($stdout =~ m{([[:xdigit:]]{32})}xsm)
    {
        $md5 = $1 ;
    }
    else
    {
        croak "no MD5 hash output?!? ($stdout)" ;
    }

    return $md5 ;
}



# Normalize by making lwercase and removing any non-hex digits.
sub get_serial_number {
    my $self = shift ;
    if (!$self->is_certificate()) {
        croak 'cannot get serial number on a non-certificate' ;
    }
    my $pem_string = $self->get_contents() ;

    my %field_to_value = parse_x509($pem_string) ;

    my $serial_number = $field_to_value{'serial'} ;

    $serial_number = lc($serial_number) ;
    $serial_number =~ s{[^abcdef0123456789]}{}gxsm ;

    return $serial_number ;
}

sub get_keysize {
    my $self = shift ;

    if ($self->is_request()) {
        return $self->get_req_keysize() ;
    } else {
        croak 'cannot get keysize for this type; ask Adam to fix' ;
    }
}

sub get_req_keysize {
    my $self = shift ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;
    close($fh) ;

    spew($self->get_contents(), $tempfile) ;

    my @cmd = (
        'openssl',
        'req',
        '-noout',
        '-in', $tempfile,
        '-inform', 'PEM',
        '-text',
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;
    $stdout = trim($stdout) ;

    my $keysize;
    if ($stdout =~ m{Key.\s+\((\d+)[ ]bit\)}ixsm) {
      $keysize = $1;
    }

    return $keysize;
}

sub get_keyinfo {
    my $self = shift ;

    if ($self->is_request()) {
        return $self->get_req_keyinfo() ;
    } else {
        croak 'cannot get keyinfo for this type; ask Adam to fix' ;
    }
}

# Returns an array with two elements:
#   ($keysize, $public_key_algorithm)
#
# $public_key_algorithm will be one of
#   * id-ecPublicKey
#   * rsaEncryption
#   * dsaEncryption
#   * hmac
#   * cmac
#   * dhKeyAgreement
#
sub get_req_keyinfo {
    my $self = shift ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    spew($self->get_contents(), $tempfile) ;

    my @cmd = (
        'openssl',
        'req',
        '-noout',
        '-in', $tempfile,
        '-inform', 'PEM',
        '-text',
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;
    $stdout = trim($stdout) ;

    my $pk_algorithm ;
    if ($stdout =~ m{Public\ Key\ Algorithm:\s+(\S+)}ixsm) {
      $pk_algorithm = $1;
    }

    my $keysize;
    if ($stdout =~ m{Key.\s+\((\d+)[ ]bit\)}ixsm) {
      $keysize = $1;
    }

    return ($keysize, $pk_algorithm) ;
}


sub is_certificate {
    my $self = shift ;
    return ($self->get_type() eq 'CERTIFICATE') ;
}

sub is_request {
    my $self = shift ;
    return ($self->get_type() eq 'REQUEST') ;
}

sub is_csr {
    my $self = shift ;
    return $self->is_request() ;
}

sub is_private_key {
    my $self = shift ;
    return ($self->get_type() eq 'PRIVKEY') ;
}

sub is_pubkey {
    my $self = shift ;
    return ($self->get_type() eq 'PUBKEY') ;
}

# Returns 1 if the public key is valid and 0 otherwise.
sub is_valid_public_key {
    my $self = shift ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;
    spew($self->get_contents(), $tempfile) ;

    my @cmd = (
        'openssl',
        'pkey',
        '-in', $tempfile,
        '-pubin',
        '-text',
        '-noout',
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;
    if ($exit_value) {
        # There was a problem, so return 0.
        return 0 ;
    } else {
        return 1 ;
    }
}

sub is_crl {
    my $self = shift ;
    return ($self->get_type() eq 'CRL') ;
}

# Will load one or more PK objects from a file. Returns an array of PKObj
# objects. If no valid PK objects in the file will return the empty array.
sub load_from_file {
    my $self = shift ;

    my $file = $self->get_file() ;

    if (!defined($file)) {
        croak 'cannot load: undefined file attribute' ;
    } elsif (!(-f $file)) {
        croak "cannot load: file '$file' not found" ;
    } else {
        my $contents = slurp($file) ;
        return get_objects_from_string($contents) ;
    }
}

# Returns the contents in DER format.
sub to_der
{
    my $self = shift ;

    my $openssl_prefix = $TYPE_TO_OPENSSL_PREFIX{$self->get_type()} ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    spew($self->get_contents(), $tempfile) ;

    my @cmd = (
        'openssl',
        $openssl_prefix,
        '-in', $tempfile,
        '-inform', 'PEM',
        '-outform', 'DER',
        ) ;
    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

    return $stdout ;
}

# Puts the DER version into the supplied file.
sub to_der_file {
    my $self = shift ;
    my ($filename_out) = @_ ;

    my $fh          = File::Temp->new(UNLINK => 1) ;
    my $tempfile_in = $fh->filename ;

    spew($self->get_contents(), $tempfile_in) ;

    my @cmd = (
        'openssl',
        'asn1parse',
        '-in',  $tempfile_in,
        '-out', $filename_out,
        '-noout',
        ) ;
    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;
    if ($stderr) {
        my $cmdx = join(q{ }, @cmd) ;
        croak "could not run command '$cmdx': $stderr" ;
    }

    return ;
}

# Parse and return a DER string. Useful for testing validity of PEM file.
sub to_der_string {
    my $self = shift ;

    my $fh_in        = File::Temp->new(UNLINK => 1) ;
    my $tempfile_in  = $fh_in->filename ;

    my $fh_out       = File::Temp->new(UNLINK => 1) ;
    my $tempfile_out = $fh_out->filename ;

    spew($self->get_contents(), $tempfile_in) ;

    my @cmd = (
        'openssl',
        'asn1parse',
        '-in',  $tempfile_in,
        '-out', $tempfile_out,
        '-noout',
        ) ;
    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;
    if ($stderr) {
        my $cmdx = join(q{ }, @cmd) ;
        croak "could not run command '$cmdx': $stderr" ;
    }

    my $der_string = slurp($tempfile_out) ;

    return $der_string ;
}


# Headers we recognize:
#
# -----BEGIN CERTIFICATE-----
# -----END CERTIFICATE-----
#
# -----BEGIN RSA PRIVATE KEY-----
# -----END RSA PRIVATE KEY-----
#
# -----BEGIN PRIVATE KEY-----
# -----END PRIVATE KEY-----
#
# -----BEGIN CERTIFICATE REQUEST-----
# -----END CERTIFICATE REQUEST-----
#
# -----BEGIN NEW CERTIFICATE REQUEST-----
# -----END NEW CERTIFICATE REQUEST-----
#
# -----BEGIN X509 CRL-----
# -----END X509 CRL-----
#
# -----BEGIN PUBLIC KEY-----
# -----END PUBLIC KEY-----
#
# Returns an array of objects.
sub get_objects_from_string
{
    my ($string) = @_ ;

    # Loop through each line
    my @objects = () ;
    while ($string) {
        ## no critic (RegularExpressions::RequireDotMatchAnything);
        ## no critic (RegularExpressions::RequireLineBoundaryMatching);
        ## no critic (RegularExpressions::RequireExtendedFormatting);
        my $type =
            qr/(?:CERTIFICATE|RSA PRIVATE KEY|PRIVATE KEY|PUBLIC KEY|CERTIFICATE REQUEST|NEW CERTIFICATE REQUEST|X509 CRL|EC PRIVATE KEY)/ ;
        my $rx = qr{(-----BEGIN ${type}-----\s)(.*?)(\s-----END ${type}-----)}s;
        if ($string =~ s{$rx}{}x) {
            my $header_line = $1 ;
            my $contents    = $1 . $2 . $3 ;

            # If any of the contents are not base64 (or dashes or spaces),
            # this is not a PEM object.
            if ($contents =~ m{[^[:alpha:][:digit:]+=\/\-\s]}ixsm) {
                return () ;
            }

            my $type1 = $HEADER_TO_TYPE{trim($header_line)} ;

            # assertion
            # uncoverable branch true
            if (!defined($type1)) {
                # uncoverable statement
                croak "could not find type corresponding to '$header_line'" ;
            }
            my $new_object = Stanford::PKObj->new() ;

            $new_object->set_type($type1) ;
            $new_object->set_encoding('PEM') ;
            $new_object->set_contents($contents) ;

            # To see if this is valid, try to compute the md5 checksum.
            eval {
                my $md5sum = $new_object->get_md5() ;
            } ;
            if ($@) {
                # There was an error. Unless this is a CRL, this must be a
                # damaged or unknown object).
                if ($type1 eq 'CRL') {
                    push(@objects, $new_object) ;
                } else {
                    # Do nothing
                }
            } else {
                push(@objects, $new_object) ;
            }
        } else {
            # No PK objects, so we are done.
            last ;
        }

        $string = trim($string) ;
    }

    return @objects ;
}

sub details
{
    my $self = shift ;

    my $openssl_prefix = $TYPE_TO_OPENSSL_PREFIX{$self->get_type()} ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    spew($self->get_contents(), $tempfile) ;

    my $pubin_option = q{} ;
    if ($self->is_pubkey) {
        $pubin_option = '-pubin' ;
    }

    my ($stdout, $stderr, $exit_value) = run_openssl_cmd(
        $openssl_prefix,
        $pubin_option,
        [
         '-in', $tempfile,
         '-text',
         '-noout',
        ],
        ) ;

    return $stdout ;
}

sub summary
{
    my $self = shift ;

    return  $self->is_certificate() ? $self->summary_certificate()
          : $self->is_request()     ? $self->summary_request()
          : $self->is_private_key() ? $self->summary_key()
          : $self->is_crl()         ? $self->summary_crl()
          : $self->is_pubkey()      ? $self->summary_key()
          :                           croak 'unrecognized type?!?'
    ;
}

sub summary_certificate
{
    my $self = shift ;

    my %field_to_value = parse_x509($self->get_contents()) ;
    my $subject = $field_to_value{'subject'} ;
    my $issuer  = $field_to_value{'issuer'} ;
    my $enddate = $field_to_value{'enddate'} ;
    my $md5     = $self->get_md5() ;

    return <<"EOC";
subject:  $subject
issuer:   $issuer
enddate:  $enddate
md5:      $md5
EOC
}

sub summary_request
{
    my $self = shift ;

    my %field_to_value = parse_req($self->get_contents()) ;

    my $subject = $field_to_value{'subject'} ;
    my $md5     = $self->get_md5() ;

    return <<"EOC";
subject:  $subject
md5:      $md5
EOC
}

sub summary_key
{
    my $self = shift ;

    my $md5 = $self->get_md5() ;

    return <<"EOC";
md5:      $md5
EOC
}

sub summary_crl
{
    my $self = shift ;

    my %field_to_value = parse_crl($self->get_contents()) ;

    my $issuer       = $field_to_value{'issuer'} ;
    my $nextupdate   = $field_to_value{'nextupdate'} ;
    my $lastupdate   = $field_to_value{'lastupdate'} ;
    my $crlnumber    = $field_to_value{'crlnumber'} ;
    my $crlnumberhex = $field_to_value{'crlnumberhex'} ;

    # Get a list of the revoked certs
    my $details_string = $self->details() ;

    my @lines  = split(/\n/xsm, $details_string) ;

    my @revoked_certs = grep { /^\s*Serial\ Number:.*$/xsm } @lines;
    my $number_of_revoked_certs = scalar(@revoked_certs) ;

    return <<"EOC";
issuer:               $issuer
nextupdate:           $nextupdate
lastupdate:           $lastupdate
crlnumber:            $crlnumber (0x$crlnumberhex)
number revoked certs: $number_of_revoked_certs
EOC
}


sub parse_x509 {
    my ($cert_string) = @_ ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    spew($cert_string, $tempfile) ;

    # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
    my $get_x509_field_func = sub
    {
      my ($fieldname, $labelname) = @_ ;
      if (!defined($labelname)) {
        $labelname = $fieldname ;
      }
      my $cmd = "openssl x509 -noout -in $tempfile -inform PEM -$fieldname" ;
      my @cmd = (
          'openssl',
          'x509',
          '-noout',
          '-in', $tempfile,
          '-inform', 'PEM',
          "-$fieldname",
          ) ;
      my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

      $stdout = trim($stdout) ;
      # uncoverable branch false
      if ($stdout =~ m{$labelname=\s*?(.*)$}xsm) {
        return trim($1) ;
      } else {
        # assertion
        # uncoverable statement
        croak "could not find field '$fieldname' in string $stdout" ;
      }
    } ;
    # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
    my %field_to_value = () ;

    $field_to_value{'subject'} = $get_x509_field_func->('subject') ;
    $field_to_value{'serial'}  = $get_x509_field_func->('serial') ;
    $field_to_value{'issuer'}  = $get_x509_field_func->('issuer') ;
    $field_to_value{'enddate'} = $get_x509_field_func->('enddate', 'notAfter') ;

    unlink $tempfile ;

    return %field_to_value ;
}

# AKA parse_csr
sub parse_req
{
    my ($req_string) = @_ ;

    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    spew($req_string, $tempfile) ;

    # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
    my $get_req_field_func = sub
    {
      my ($fieldname, $labelname) = @_ ;
      # uncoverable branch false
      if (!defined($labelname)) {
        $labelname = $fieldname ;
      }

      my @cmd = (
          'openssl',
          'req',
          '-noout',
          '-in', $tempfile,
          '-inform', 'PEM',
          "-$fieldname",
          ) ;
      my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

      $stdout = trim($stdout) ;
      # uncoverable branch false
      if ($stdout =~ m{$labelname=\s*?(.*)$}xsm) {
        return trim($1) ;
      } else {
        # assertion
        # uncoverable statement
        croak "could not find field '$fieldname' in string $stdout" ;
      }
    } ;
    # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
    my %field_to_value = () ;

    $field_to_value{'subject'} = $get_req_field_func->('subject') ;

    return %field_to_value ;
}

sub parse_crl
{
  my ($crl_string) = @_ ;

  my $fh       = File::Temp->new(UNLINK => 1) ;
  my $tempfile = $fh->filename ;

  spew($crl_string, $tempfile) ;

  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
  my $get_crl_field_func = sub
  {
    my ($fieldname, $labelname) = @_ ;
    if (!defined($labelname)) {
      $labelname = $fieldname ;
    }

    my @cmd = (
        'openssl',
        'crl',
        '-noout',
        '-in', $tempfile,
        '-inform', 'PEM',
        "-$fieldname",
        );
    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

    $stdout = trim($stdout) ;
    if ($fieldname eq 'hash') {
      return trim($stdout) ;
    }
    # uncoverable branch false
    elsif ($stdout =~ m{$labelname=\s*?(.*)$}xsm) {
      return trim($1) ;
    } else {
      # assertion
      # uncoverable statement
      croak "could not find field '$fieldname' in string '$stdout'" ;
    }
  } ;
  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
  my %field_to_value = () ;

  $field_to_value{'fingerprint'} = $get_crl_field_func->('fingerprint', 'SHA1\ Fingerprint') ;
  $field_to_value{'issuer'}      = $get_crl_field_func->('issuer') ;
  $field_to_value{'lastupdate'}  = $get_crl_field_func->('lastupdate', 'lastUpdate') ;
  $field_to_value{'nextupdate'}  = $get_crl_field_func->('nextupdate', 'nextUpdate') ;
  $field_to_value{'crlnumberhex'}= $get_crl_field_func->('crlnumber', 'crlNumber') ;
  $field_to_value{'hash'}        = $get_crl_field_func->('hash', q{}) ;

  $field_to_value{'crlnumber'}   = hex($field_to_value{'crlnumberhex'}) ;

  unlink $tempfile ;

  return %field_to_value ;
}

# Try to guess what sort of object it is and return the parsed object if
# we guess correctly.
#
# Try different ways to read a PKObj.
# 1. sandwich string with headers
# 2. maybe it's a DER object?
sub get_object_by_guessing {
    my ($string) = @_ ;

    $string = trim($string) ;

    my @objects ;

    # 1. Try adding header/footer.
    my @headers_to_try = (
        'CERTIFICATE',
        'RSA PRIVATE KEY',
        'PRIVATE KEY',
        'CERTIFICATE REQUEST',
        'NEW CERTIFICATE REQUEST',
        'X509 CRL',
        'PUBLIC KEY',
        ) ;

    foreach my $header (@headers_to_try) {
        my $s = "-----BEGIN $header-----" . "\n"
              . $string                   . "\n"
              . "-----END $header-----" ;

        @objects = get_objects_from_string($s) ;
        if (scalar(@objects) > 0) {
            my $object = $objects[0] ;
            my $md5 ;
            eval {
                $md5 = $object->get_md5() ;
            } ;
            if ($md5) {
                return $object ;
            }
        }
    }

    # 2. Maybe it's a DER string.
    my $fh       = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $fh->filename ;

    # Try DER
    spew($string, $tempfile) ;
    foreach my $type (@ALLOWED_TYPES) {
        my $openssl_prefix = $TYPE_TO_OPENSSL_PREFIX{$type} ;

        my @cmd = (
            'openssl',
            $openssl_prefix,
            '-noout',
            '-in', $tempfile,
            '-inform', 'DER',
            ) ;
        my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

        if ($exit_value == 0) {
            # This must be it!
            # Convert to base64, sandwich the appropriate headers, and pass to
            # get_objects_from_string.

            my $base64_string = encode_base64($string);
            my $header = $TYPE_TO_HEADER{$type} ;
            my $footer = $TYPE_TO_FOOTER{$type} ;
            my $pem_string = $header . "\n" . $base64_string . $footer . "\n" ;

            @objects = get_objects_from_string($pem_string) ;
            return $objects[0] ;
        }
    }

    return ;
}

# Returns undef if cannot extract.
sub get_CN_from_subject { ## no critic (NamingConventions::Capitalization);
    my ($subject) = @_ ;

    if (!$subject) {
        return ;
    } else {
        if ($subject =~ m{CN=(.*?)([,\/]|$)}ixsm) {
            return $1 ;
        } else {
            return ;
        }
    }
}

sub run_openssl_cmd {
    my ($prefix, $pubin_option, $other_args_aref) = @_ ;

    my @cmd ;

    if ($pubin_option) {
        @cmd = (
            'openssl',
            $prefix,
            $pubin_option,
            @{ $other_args_aref },
            );
    } else {
        @cmd = (
            'openssl',
            $prefix,
            @{ $other_args_aref },
            ) ;
    } ;

    return run_command_improved(@cmd) ;
}


1 ;

__END__


=head1 NAME

Stanford::PKObj - Base class for certificates, private keys, CSRs, etc.

=head1 VERSION

Version 0.01

=head1 ABSTRACT

The Stanford::PKObj is the base class that classes such as Stanford::CSR,
Stanford::x509, and Stanford::RSAKey extend.

=head1 AUTHOR

Adam Lewenberg, C<< <adamhl at stanford.edu> >>

=head1 BUGS

Please report any bugs or feature requests
to C<adamhl at stanford.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Stanford::PKObj

=head1 COPYRIGHT & LICENSE

Copyright 2012 Stanford University, all rights reserved.

=cut
